/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPTIMIZER_ADAM_H_
#define AIDGE_CORE_OPTIMIZER_ADAM_H_

#include <functional>
#include <memory>
#include <vector>
#include <cmath>  // std::sqrt, std::pow

#include "aidge/data/Tensor.hpp"
#include "aidge/learning/optimizer/Optimizer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/TensorUtils.hpp"

namespace Aidge {

enum class AdamAttr {
    Beta1,
    Beta2,
    Epsilon
};

class Adam: public Optimizer, public StaticAttributes<AdamAttr, float, float, float> {
private:
    std::vector<Tensor> mMomentum1;
    std::vector<Tensor> mMomentum2;
    Tensor mBeta1;
    Tensor mReversedBeta1;
    Tensor mBeta2;
    Tensor mReversedBeta2;

public:
    using Attributes_ = StaticAttributes<AdamAttr, float, float, float>;
    template <AdamAttr e>
    using attr = typename Attributes_::template attr<e>;

    Adam(const float beta1 = 0.9f, const float beta2 = 0.999f, const float epsilon = 1.0e-8f)
        : Optimizer(),
          Attributes_(attr<AdamAttr::Beta1>(beta1),
                      attr<AdamAttr::Beta2>(beta2),
                      attr<AdamAttr::Epsilon>(epsilon))
    {
        mBeta1 = Tensor(beta1);
        mReversedBeta1 = Tensor(1.0f - beta1);
        mBeta2 = Tensor(beta2);
        mReversedBeta2 = Tensor(1.0f - beta2);
    }

    void update() override final {
        float mBeta1Power = std::pow(this->getAttr<AdamAttr::Beta1>(), static_cast<float>(mLRScheduler.step() + 1));
        float mBeta2Power = std::pow(this->getAttr<AdamAttr::Beta2>(), static_cast<float>(mLRScheduler.step() + 1));
        float mReversedBeta1Power = 1.0f - mBeta1Power;
        float mSqrtReversedBeta2Power = std::sqrt(1.0f - mBeta2Power);

        Tensor alpha = Tensor(learningRate() * mSqrtReversedBeta2Power / mReversedBeta1Power);
        alpha.setBackend(mParameters[0]->getImpl()->backend());
        alpha.setDataType(mParameters[0]->dataType());

        Tensor epsilon_hat = Tensor(this->getAttr<AdamAttr::Epsilon>() * mSqrtReversedBeta2Power);
        epsilon_hat.setBackend(mParameters[0]->getImpl()->backend());
        epsilon_hat.setDataType(mParameters[0]->dataType());

        if (mLRScheduler.step() == 0) {
            for (std::size_t i = 0; i < mParameters.size(); ++i) {
                mMomentum1[i].zeros();
                mMomentum2[i].zeros();
            }
        }

        for (std::size_t i = 0; i < mParameters.size(); ++i) {
            mMomentum1[i] = mBeta1 * mMomentum1[i] + mReversedBeta1 * (*mParameters[i]->grad());
            mMomentum2[i] = mBeta2 * mMomentum2[i] + mReversedBeta2 * (*mParameters[i]->grad()) * (*mParameters[i]->grad());
            *mParameters[i] -= alpha * mMomentum1[i] / (mMomentum2[i].sqrt() + epsilon_hat);
        }

        mLRScheduler.update();
    }

    void setParameters(const std::vector<std::shared_ptr<Tensor>>& parameters) override final {
        Optimizer::setParameters(parameters);
        mMomentum1 = std::vector<Tensor>(parameters.size());
        mMomentum2 = std::vector<Tensor>(parameters.size());
        for (std::size_t i = 0; i < parameters.size(); ++i) {
            mMomentum1[i] = Tensor(parameters[i]->dims());
            mMomentum1[i].setBackend(parameters[i]->getImpl()->backend());
            mMomentum1[i].setDataType(parameters[i]->dataType());
            mMomentum2[i] = Tensor(parameters[i]->dims());
            mMomentum2[i].setBackend(parameters[i]->getImpl()->backend());
            mMomentum2[i].setDataType(parameters[i]->dataType());
        }
        if (parameters.size() > 0) {
            mBeta1.setBackend(mParameters[0]->getImpl()->backend());
            mBeta1.setDataType(parameters[0]->dataType());
            mReversedBeta1.setBackend(mParameters[0]->getImpl()->backend());
            mReversedBeta1.setDataType(parameters[0]->dataType());
            mBeta2.setBackend(mParameters[0]->getImpl()->backend());
            mBeta2.setDataType(parameters[0]->dataType());
            mReversedBeta2.setBackend(mParameters[0]->getImpl()->backend());
            mReversedBeta2.setDataType(parameters[0]->dataType());
        }
    }
};

} // namespace Aidge


namespace {
template <>
const char *const EnumStrings<Aidge::AdamAttr>::data[] = {
    "Beta1",
    "Beta2",
    "Epsilon"
};
}
#endif // AIDGE_CORE_OPTIMIZER_ADAM_H_
