/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <catch2/catch_test_macros.hpp>
#include <cstddef>     // std::size_t
#include <cmath>       //
#include <functional>  // std::multiplies, std::plus
#include <memory>      // std::make_unique
#include <numeric>     // std::accumulate
#include <random>      // std::random_device, std::mt19937,
                       // std::uniform_int_distribution
#include <vector>

#include "aidge/loss/LossList.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/TensorUtils.hpp"

#include "aidge/backend/cpu/operator/SoftmaxImpl.hpp"
#if USE_AIDGE_BACKEND_CUDA
#include "aidge/backend/cuda/operator/SoftmaxImpl.hpp"
#endif

namespace Aidge {

// Utility that compute the CELoss manually

static float manualCELoss(float *predictionArray, float *targetArray, std::size_t batchSize, std::size_t outputSize)
{
    const std::size_t nbElements = batchSize * outputSize;

    float *softmaxArray = new float[nbElements];
    for (std::size_t i = 0; i < batchSize; ++i) {
        float partition = 0;
        for (std::size_t j = 0; j < outputSize; ++j) {
            float value = std::exp(predictionArray[i * outputSize + j]);
            softmaxArray[i * outputSize + j] = value;
            partition += value;
        }
        for (std::size_t j = 0; j < outputSize; ++j) {
            softmaxArray[i * outputSize + j] /= partition;
        }
    }

    float* productArray = new float[nbElements];
    for (std::size_t i = 0; i < nbElements; ++i)
        productArray[i] = targetArray[i] * std::log(softmaxArray[i]);

    float* sumArray = new float[batchSize];
    for (std::size_t i = 0; i < batchSize; ++i) {
        float acc = 0;
        for (std::size_t j = 0; j < outputSize; ++j)
            acc += productArray[i * outputSize + j];
        sumArray[i] = acc;
    }
        
    float mean = 0;
    for (std::size_t i = 0; i < batchSize; ++i)
        mean += sumArray[i] / static_cast<float> (batchSize);

    delete[] softmaxArray; 
    delete[] productArray;      
    delete[] sumArray;      

    return -mean;
}

TEST_CASE("[loss/classification] CELoss", "[loss][classification][CELoss]") {

    // CONSTANTS

    constexpr std::uint16_t NB_TRIALS = 10;

    // SETUP RNGS

    std::mt19937 gen(100);
    std::uniform_real_distribution<float> valueDist(-2, 2);
    std::uniform_int_distribution<std::size_t> dimSizeDist(std::size_t(1), std::size_t(8));

    SECTION("CPU") {

        for (std::uint16_t trial = 0; trial < NB_TRIALS; ++trial) 
        {
            const std::size_t nbDims = 2;
            const std::size_t batchSize = dimSizeDist(gen);
            const std::size_t outputSize = dimSizeDist(gen);

            std::vector<std::size_t> dims;
            dims.push_back(batchSize);
            dims.push_back(outputSize);

            const std::size_t nbElements = batchSize * outputSize;

            // Create the data array/tensors

            float* predictionArray = new float[nbElements];
            for (std::size_t i = 0; i < nbElements; ++i)
                predictionArray[i] = valueDist(gen);

            float* targetArray = new float[nbElements];
            for (std::size_t i = 0; i < nbElements; ++i)
                targetArray[i] = valueDist(gen);

            std::shared_ptr<Tensor> predictionTensor = std::make_shared<Tensor>(dims);
            predictionTensor->setBackend("cpu");
            predictionTensor->setDataType(DataType::Float32);
            predictionTensor->getImpl()->setRawPtr(predictionArray, nbElements);

            std::shared_ptr<Tensor> targetTensor = std::make_shared<Tensor>(dims);
            targetTensor->setBackend("cpu");
            targetTensor->setDataType(DataType::Float32);
            targetTensor->getImpl()->setRawPtr(targetArray, nbElements);

            // Compute the CELoss manually

            Tensor manualResult = Tensor(manualCELoss(predictionArray, targetArray, batchSize, outputSize));
            manualResult.resize({1, 1});
          
            // Compute the CELoss using Aidge::loss::CELoss function

            const Tensor functionResult = loss::CELoss(predictionTensor, targetTensor);

            // Compare results

            Log::info( " CELoss = {} {} ", manualResult.get<float>(0), functionResult.get<float>(0));
            REQUIRE(approxEq<float>(manualResult, functionResult));   

            // Free memory

            delete[] predictionArray; 
            delete[] targetArray;               
        }
    }
#if USE_AIDGE_BACKEND_CUDA
    SECTION("CUDA") {

        for (std::uint16_t trial = 0; trial < NB_TRIALS; ++trial) 
        {
            const std::size_t nbDims = 2;
            const std::size_t batchSize = dimSizeDist(gen);
            const std::size_t outputSize = dimSizeDist(gen);

            std::vector<std::size_t> dims;
            dims.push_back(batchSize);
            dims.push_back(outputSize);

            const std::size_t nbElements = batchSize * outputSize;

            // Create the arrays/tensors

            float* predictionArray = new float[nbElements];
            for (std::size_t i = 0; i < nbElements; ++i)
                predictionArray[i] = valueDist(gen);

            float* targetArray = new float[nbElements];
            for (std::size_t i = 0; i < nbElements; ++i)
                targetArray[i] = valueDist(gen);

            std::shared_ptr<Tensor> predictionTensor = std::make_shared<Tensor>(dims);
            predictionTensor->setDataType(DataType::Float32);
            predictionTensor->setBackend("cuda");

            float* predictionArrayDevice;
            cudaMalloc(reinterpret_cast<void **> (&predictionArrayDevice), sizeof(float) * nbElements);
            cudaMemcpy(predictionArrayDevice, predictionArray, sizeof(float) * nbElements, cudaMemcpyHostToDevice);
            predictionTensor->getImpl()->setRawPtr(predictionArrayDevice, nbElements);

            std::shared_ptr<Tensor> targetTensor = std::make_shared<Tensor>(dims);
            targetTensor->setDataType(DataType::Float32);
            targetTensor->setBackend("cuda");

            float* targetArrayDevice;
            cudaMalloc(reinterpret_cast<void **> (&targetArrayDevice), sizeof(float) * nbElements);
            cudaMemcpy(targetArrayDevice, targetArray, sizeof(float) * nbElements, cudaMemcpyHostToDevice);
            targetTensor->getImpl()->setRawPtr(targetArrayDevice, nbElements);

            // Compute the CELoss manually

            Tensor manualResult = Tensor(manualCELoss(predictionArray, targetArray, batchSize, outputSize));
            manualResult.resize({1, 1});
          
            // Compute the CELoss using Aidge::loss::CELoss function

            const Tensor functionResult = loss::CELoss(predictionTensor, targetTensor);

            // Compare results

            Log::info(" CELoss = {} {} ", manualResult.get<float>(0), functionResult.get<float>(0));
            REQUIRE(approxEq<float>(manualResult, functionResult));   

            // Free memory

            delete[] predictionArray; 
            delete[] targetArray;
            cudaFree(predictionArrayDevice);
            cudaFree(targetArrayDevice);
        }
    }
#endif
}
}  // namespace Aidge
